SystemOrganization addCategory: #'Smallapack-OpalForLegacyFFI'!

!RBToken methodsFor: '*Smallapack-OpalForLegacyFFI' stamp: 'nice 1/10/2019 16:54'!
isLegacyFFIKeyword
	^false! !

!RBKeywordToken methodsFor: '*Smallapack-OpalForLegacyFFI' stamp: 'nice 1/10/2019 16:55'!
isLegacyFFIKeyword
	^ self value = #cdecl: or: [ self value = #apicall: ]! !

!SLASTTranslator methodsFor: '*Smallapack-OpalForLegacyFFI' stamp: 'nice 1/10/2019 23:39'!
decodeExternalTypeFrom: tokenStream descriptorClass: descriptorClass
	| tmp xType |
	tmp := tokenStream peek.
	xType := descriptorClass atomicTypeNamed: tmp.
	xType == nil ifTrue:["Look up from class scope"
		Symbol hasInterned: tmp ifTrue:[:sym|
			xType := descriptorClass structTypeNamed: sym]].
	xType ifNil: [
		"Error, no such type"
		^nil].
	tokenStream next.
	(tokenStream peekFor: #*)
		ifTrue:[^xType asPointerType]
		ifFalse:[^xType]! !

!SLASTTranslator methodsFor: '*Smallapack-OpalForLegacyFFI' stamp: 'nice 1/11/2019 00:14'!
decodeFunctionDeclarationPragma: aPragmaNode methodNode: aMethodNode
	"Parse the function declaration for a call to an external library."

	| descriptorClass callType retType externalName args argType module keys tokenStream externalArgs |
	descriptorClass := (Smalltalk globals
		at: #ExternalFunction
		ifAbsent: [ nil ]) ifNil: [ ^ nil ].
	aPragmaNode selector isKeyword
		ifFalse: [ ^ nil ].
	keys := aPragmaNode selector keywords.
	callType := (descriptorClass callingConventionFor: keys first)
		ifNil: [ ^ nil ].	"Parse return type"
	tokenStream := aPragmaNode arguments first value readStream.
	retType := (self
		decodeExternalTypeFrom: tokenStream
		descriptorClass: descriptorClass)
		ifNil: [ "^ self expected: 'return type' " ^ nil ].

	"Parse function name or index"
	externalName := tokenStream next.
	externalName isString
		ifTrue: [ externalName := externalName asSymbol ]
		ifFalse: [ externalName isInteger
				ifFalse: [ "^ self expected: 'function name or index'" ^ nil ] ].
	externalArgs := tokenStream next.
	externalArgs isArray
		ifFalse: [ "^ self expected: 'argument list'" ^ nil ].
	externalArgs := externalArgs readStream.
	args := Array new writeStream.
	[ externalArgs atEnd ]
		whileFalse: [ argType := (self
				decodeExternalTypeFrom: externalArgs
				descriptorClass: descriptorClass)
				ifNil: [ "^ self expected: 'argument'" ^ nil ].
			argType isVoid & argType isPointerType not
				ifFalse: [ args nextPut: argType ] ].
	args position = aMethodNode arguments size
		ifFalse: [ "^self expected: 'Matching number of arguments'" ^ nil ].
	(keys size = 2 and: [ (keys at: 2) = 'module:' ])
		ifTrue: [ module := aPragmaNode arguments at: 2.
			(module isLiteralNode and: [ module value isString ])
				ifFalse: [ "^ self expected: 'String'" ^ nil ].
			module := module value asSymbol ].
	^ Smalltalk globals
		at: #ExternalLibraryFunction
		ifPresent: [ :elf | 
			elf
				name: externalName
				module: module
				callType: callType
				returnType: retType
				argumentTypes: args contents ]! !

!SLASTTranslator methodsFor: '*Smallapack-OpalForLegacyFFI' stamp: 'nice 1/11/2019 00:58'!
visitPragmaNode: aPragma
	
	"horrible hack"
	| aMethodNode |
	aMethodNode := thisContext sender home tempAt: 1.
	
	(self decodeFunctionDeclarationPragma: aPragma methodNode: aMethodNode)
		ifNil: [ super visitPragmaNode: aPragma ]
		ifNotNil:
			[ :fn |
			methodBuilder irPrimitive: (IRPrimitive new
				num: 120;
				spec: fn;
				yourself)].! !

!SLCompilationContext methodsFor: '*Smallapack-OpalForLegacyFFI' stamp: 'nice 1/10/2019 16:34'!
parserClass
	^ parserClass ifNil: [ parserClass := SLParser ]! !

RBParser subclass: #SLParser
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Smallapack-OpalForLegacyFFI'!

!SLParser commentStamp: 'nice 1/10/2019 17:32' prior: 0!
I am a Parser for  making the Smallapack Opal Compiler able to compile legacy FFI pragmas.

I should be used only if such legacy FFI  pragmas are still in use in Smallapack source code.

Starting at Pharo version 6, UnifiedFFI and #ffiCall: syntax should be preferred.
I exist only for making the transition easier in the interim and should better not be used in Pharo version 7 and later, unless for facilitating the import of Squeak source code.

Note that I have nothing specific to Smallapack, so I could be used in other projects than Smallapack too, for backward compatibility and for facilitating the port to UnifiedFFI.!

!SLParser methodsFor: 'private-parsing' stamp: 'nice 1/10/2019 16:56'!
basicParsePragma
	^ currentToken isIdentifier
		ifTrue: [ self parseUnaryPragma ]
		ifFalse: [
			currentToken isKeyword
				ifTrue: [ 	
					currentToken isPrimitiveKeyword
						ifTrue: [ ^ self parsePrimitiveKeywordPragma ].
					currentToken isLegacyFFIKeyword
						ifTrue: [ ^ self parseLegacyFFIPragma ].
					self parseKeywordPragma ]
				ifFalse: [ self parseBinaryPragma ] ]! !

!SLParser methodsFor: 'private-parsing' stamp: 'nice 1/10/2019 23:27'!
parseLegacyFFIPragma
	| selector positions arguments spec argStart argStop fakeLiteralArray |
	selector := currentToken value.
	positions := (Array new: 5) writeStream.
	arguments := (Array new: 5) writeStream.
	spec := (Array new: 5) writeStream.
	
	positions nextPut: currentToken start.
	self step.
	argStop := argStart := currentToken start .
	
	[ ( currentToken isBinary and: [ currentToken value == #> ] )
			or: [ currentToken isKeyword ] ]
		 whileFalse: 
			[argStop := currentToken stop.
			spec nextPut: self parseLiteralArrayObject].
	fakeLiteralArray := self literalArrayNodeClass 
		startPosition: argStart
		contents: spec contents
		stopPosition: argStop
		isByteArray: false.
	arguments nextPut: fakeLiteralArray.
	
	"eventually parse module: specification"
	[ currentToken isKeyword ] whileTrue: [
		selector := selector, currentToken value.
		positions nextPut: currentToken start.
		self step.
		arguments nextPut: self parsePragmaLiteral ].
	
	^ self pragmaNodeClass
		selector: selector asSymbol
		keywordsPositions: positions contents asIntegerArray
		arguments: arguments contents.! !
